@@ -1,16 +1,16 @@
 /* revision.h -- define the version number
+ * Copyright (C) 1998, 1999, 2001 Free Software Foundation, Inc.
  * Copyright (C) 1992-1993 Jean-loup Gailly.
  * This is free software; you can redistribute it and/or modify it under the
  * terms of the GNU General Public License, see the file COPYING.
  */
 
-#define VERSION "1.2.3"
 #define PATCHLEVEL 0
-#define REVDATE "24 Jun 93"
+#define REVDATE "2001-11-03"
 
 /* This version does not support compression into old compress format: */
 #ifdef LZW
 #  undef LZW
 #endif
 
-/* $Id: revision.h,v 0.24 1993/06/17 11:48:02 jloup Exp $ */
+/* $Id: revision.h,v 0.25 1993/06/24 08:29:52 jloup Exp $ */
